/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef CSR_MATRIX_DATA_GPU_H
#define CSR_MATRIX_DATA_GPU_H

// -----------------------------------------------------------------------------

#include "csr_matrix_data.h"
#include "cuda_memory.h"

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
class csr_matrix_data_gpu : public csr_matrix_data<index_type, value_type>
{
public:

	typedef csr_matrix_data<index_type, value_type> csr_matrix_data_type;

	csr_matrix_data_gpu()
		: csr_matrix_data<index_type, value_type>()
	{
	}

	virtual ~csr_matrix_data_gpu()
	{
	}

	bool copy_from_host(const csr_matrix_data<index_type, value_type>& cpu_data);

	// just copying: space for data on host will not be allocated.
	bool copy_to_host(csr_matrix_data<index_type, value_type>& cpu_data) const;

	virtual bool clear();

private:

	cuda_memory<index_type> gpu_row_offsets;
	cuda_memory<index_type> gpu_column_indices;
	cuda_memory<value_type> gpu_values;

private:
	
	csr_matrix_data_gpu(const csr_matrix_data_gpu&);
	csr_matrix_data_gpu& operator=(const csr_matrix_data_gpu&);
};

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_matrix_data_gpu<index_type, value_type>::copy_from_host(
	const csr_matrix_data<index_type, value_type>& cpu_data)
{
	bool success = false;

	if(cpu_data.num_rows > 0 && cpu_data.num_cols > 0)
	{
		success = gpu_row_offsets.resize   (cpu_data.num_rows + 1, cpu_data.row_offsets) &&
		          gpu_column_indices.resize(cpu_data.num_entries, cpu_data.column_indices) &&
		          gpu_values.resize        (cpu_data.num_entries, cpu_data.values);

		if(success)
		{
			csr_matrix_data_type::num_rows       = cpu_data.num_rows;
			csr_matrix_data_type::num_cols       = cpu_data.num_cols;
			csr_matrix_data_type::num_entries    = cpu_data.num_entries;

			csr_matrix_data_type::row_offsets    = gpu_row_offsets.gpu_mem_ptr();
			csr_matrix_data_type::column_indices = gpu_column_indices.gpu_mem_ptr();
			csr_matrix_data_type::values         = gpu_values.gpu_mem_ptr();
		}
	}

	return success;
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_matrix_data_gpu<index_type, value_type>::copy_to_host(
	csr_matrix_data<index_type, value_type>& cpu_data) const
{
	bool success = false;

	if( cpu_data.num_rows == csr_matrix_data_type::num_rows &&
		cpu_data.num_cols == csr_matrix_data_type::num_cols &&
		cpu_data.num_entries == csr_matrix_data_type::num_entries &&
		cpu_data.row_offsets &&
		cpu_data.column_indices &&
		cpu_data.values &&
		csr_matrix_data_type::row_offsets &&
		csr_matrix_data_type::column_indices &&
		csr_matrix_data_type::values)
	{
		success = gpu_row_offsets.copy_to_host   (cpu_data.row_offsets) &&
		          gpu_column_indices.copy_to_host(cpu_data.column_indices) &&
		          gpu_values.copy_to_host        (cpu_data.values);
	}

	return success;
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_matrix_data_gpu<index_type, value_type>::clear()
{
	bool success =
	    csr_matrix_data_type::clear() &&
	    gpu_row_offsets.clear() &&
	    gpu_column_indices.clear() &&
	    gpu_values.clear();

	return success;
}

// -----------------------------------------------------------------------------

#endif // CSR_MATRIX_DATA_GPU_H
