/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef CSR_SPMV_INODE_H
#define CSR_SPMV_INODE_H

// -----------------------------------------------------------------------------

#include "csr_matrix_data.h"
#include "operator_T.h"
#include <cstddef>
#include <stdio.h>
#include <iostream>
#include <algorithm> // std::min
#include <cuda.h>
#ifdef CUSP_USE_TEXTURE_MEMORY
#include <cusp/detail/device/texture.h>
#endif
#include <thrust/detail/device/cuda/arch.h>

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_csr_vector_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2010 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

template <typename index_type, typename value_type, int UseCache, unsigned int THREADS_PER_VECTOR, unsigned int NNZ_PER_ROW_MAX, unsigned int NODEMAX>
__global__ void csr_spmv_inode_kernel(const index_type numNodes,
				      const index_type nodeMax,
				      const index_type * Ap, // A.row_offsets
				      const index_type * Aj, // A.column_indices
				      const value_type * Ax, // A.values
				      const index_type * inodes,
				      const value_type * x,
				      value_type* y)   
{
  __shared__ volatile value_type xdata[NNZ_PER_ROW_MAX];
  __shared__ volatile value_type sdata[NODEMAX*THREADS_PER_VECTOR + THREADS_PER_VECTOR/2];
  __shared__ volatile index_type row_offsets[2];
  __shared__ volatile index_type inode0;
  __shared__ volatile index_type inode1;

  /* thread id in the inode block */
  const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);
    
  /* vector index within the block */
  const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;

  /* index within the shared memory data structure for the reduction */
  const index_type shared_mem_lane = vector_lane * THREADS_PER_VECTOR + thread_lane;

  for(index_type row = blockIdx.x; row < numNodes; row += gridDim.x) {
    /* Get the inode and the row offsets */
    if (threadIdx.x==0) inode0 = inodes[row];
    if (threadIdx.x==1) inode1 = inodes[row+1];
    __syncthreads();

    index_type inode = inode0;
    if(threadIdx.x < 2) row_offsets[threadIdx.x] = Ap[inode+threadIdx.x];
    __syncthreads();

    index_type row_start  = row_offsets[0];
    index_type row_diff   = row_offsets[1]-row_start;

    /* next, load x into shared memory, xdata */
    if (threadIdx.x < row_diff)
      xdata[threadIdx.x] = fetch_x<UseCache> (Aj[row_start + threadIdx.x], x);
    __syncthreads();

    if (vector_lane < inode1-inode) {
      /* initialize local sum */
      value_type sum = 0.;

      /* accumulate local sums */
      index_type row_shift  = row_start + vector_lane*row_diff;
      index_type jj = thread_lane;
      if(jj < row_diff)
        sum += Ax[jj + row_shift] * xdata[jj];
      for(jj += THREADS_PER_VECTOR; jj < row_diff; jj += THREADS_PER_VECTOR)
        sum += Ax[jj + row_shift] * xdata[jj];

      /* store local sum in shared memory */
      sdata[shared_mem_lane] = sum;

      /* reduce local sums to row sum */
      if(THREADS_PER_VECTOR > 16) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane + 16];
      if(THREADS_PER_VECTOR >  8) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  8];
      if(THREADS_PER_VECTOR >  4) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  4];
      if(THREADS_PER_VECTOR >  2) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  2];
      if(THREADS_PER_VECTOR >  1) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  1];

      /* first thread of EACH vector writes the result */
      if(thread_lane==0)
        y[inode + vector_lane] = sum;
      
    }
  }
}


// -----------------------------------------------------------------------------


template <int UseCache, unsigned int THREADS_PER_VECTOR, unsigned int NNZ_PER_ROW_MAX, typename index_type, typename value_type>
static void csr_spmv_inode_T(const csr_matrix_data<index_type, value_type>& A, 
			     const index_type numNodes, 
			     const index_type nodeMax,
			     const index_type* inodes, 
			     const value_type* x,
			     value_type* y)
{
  if (UseCache)
    bind_x(x);

  if (nodeMax==1) {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_inode_kernel<index_type, value_type, USECACHE, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 1>, THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    csr_spmv_inode_kernel<index_type, value_type, UseCache, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 1>
      <<<NUM_BLOCKS, THREADS_PER_VECTOR>>>(numNodes, nodeMax,
                                           A.row_offsets,
                                           A.column_indices,
                                           A.values,
                                           inodes, x, y);
  }
  else if (nodeMax==2) {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_inode_kernel<index_type, value_type, USECACHE, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 2>, 2*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    csr_spmv_inode_kernel<index_type, value_type, UseCache, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 2>
      <<<NUM_BLOCKS, 2*THREADS_PER_VECTOR>>>(numNodes, nodeMax,
                                             A.row_offsets,
                                             A.column_indices,
                                             A.values,
                                             inodes, x, y);
  }
  else if (nodeMax==3) {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_inode_kernel<index_type, value_type, USECACHE, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 3>, 3*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    csr_spmv_inode_kernel<index_type, value_type, UseCache, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 3>
      <<<NUM_BLOCKS, 3*THREADS_PER_VECTOR>>>(numNodes, nodeMax,
                                             A.row_offsets,
                                             A.column_indices,
                                             A.values,
                                             inodes, x, y);
  }
  else if (nodeMax==4) {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_inode_kernel<index_type, value_type, USECACHE, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 4>, 4*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    csr_spmv_inode_kernel<index_type, value_type, UseCache, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 4>
      <<<NUM_BLOCKS, 4*THREADS_PER_VECTOR>>>(numNodes, nodeMax,
                                             A.row_offsets,
                                             A.column_indices,
                                             A.values,
                                             inodes, x, y);
  }
  else if (nodeMax==5) {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_inode_kernel<index_type, value_type, USECACHE, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 5>, 5*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);
    csr_spmv_inode_kernel<index_type, value_type, UseCache, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 5>
      <<<NUM_BLOCKS, 5*THREADS_PER_VECTOR>>>(numNodes, nodeMax,
                                             A.row_offsets,
                                             A.column_indices,
                                             A.values,
                                             inodes, x, y);
  }

  if (UseCache)
    unbind_x(x);
}


// -----------------------------------------------------------------------------

template <typename index_type, typename value_type>
int csr_spmv_inode(const csr_matrix_data<index_type, value_type>* A, 
		   const index_type numNodes,
		   const index_type nodeMax,
		   const index_type nnzPerRowMax,
		   const index_type* inodes, 
		   const value_type* x,
		   value_type* y)
{
  const index_type nnz_per_row = A->num_entries / A->num_rows;

#ifdef CUSP_USE_TEXTURE_MEMORY 
  #define USECACHE 1
#else
  #define USECACHE 0
#endif

  if(nnz_per_row <=  2) {
    if (nnzPerRowMax <=  2)    { csr_spmv_inode_T<USECACHE, 2,   2>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  4)    { csr_spmv_inode_T<USECACHE, 2,   4>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  8)    { csr_spmv_inode_T<USECACHE, 2,   8>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  16)   { csr_spmv_inode_T<USECACHE, 2,  16>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  32)   { csr_spmv_inode_T<USECACHE, 2,  32>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  64)   { csr_spmv_inode_T<USECACHE, 2,  64>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  128)  { csr_spmv_inode_T<USECACHE, 2, 128>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  256)  { csr_spmv_inode_T<USECACHE, 2, 256>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  512)  { csr_spmv_inode_T<USECACHE, 2, 512>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  1024) { csr_spmv_inode_T<USECACHE, 2,1024>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { csr_spmv_inode_T<USECACHE, 2,2048>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#endif
  }
  if(nnz_per_row <=  4) {
    if (nnzPerRowMax <=  4)    { csr_spmv_inode_T<USECACHE, 4,   4>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  8)    { csr_spmv_inode_T<USECACHE, 4,   8>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  16)   { csr_spmv_inode_T<USECACHE, 4,  16>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  32)   { csr_spmv_inode_T<USECACHE, 4,  32>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  64)   { csr_spmv_inode_T<USECACHE, 4,  64>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  128)  { csr_spmv_inode_T<USECACHE, 4, 128>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  256)  { csr_spmv_inode_T<USECACHE, 4, 256>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  512)  { csr_spmv_inode_T<USECACHE, 4, 512>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  1024) { csr_spmv_inode_T<USECACHE, 4,1024>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { csr_spmv_inode_T<USECACHE, 4,2048>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#endif
  }
  if(nnz_per_row <=  8) {
    if (nnzPerRowMax <=  8)    { csr_spmv_inode_T<USECACHE, 8,   8>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  16)   { csr_spmv_inode_T<USECACHE, 8,  16>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  32)   { csr_spmv_inode_T<USECACHE, 8,  32>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  64)   { csr_spmv_inode_T<USECACHE, 8,  64>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  128)  { csr_spmv_inode_T<USECACHE, 8, 128>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  256)  { csr_spmv_inode_T<USECACHE, 8, 256>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  512)  { csr_spmv_inode_T<USECACHE, 8, 512>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  1024) { csr_spmv_inode_T<USECACHE, 8,1024>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { csr_spmv_inode_T<USECACHE, 8,2048>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#endif
  }
  if(nnz_per_row <= 16) {
    if (nnzPerRowMax <=  16)   { csr_spmv_inode_T<USECACHE,16,  16>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  32)   { csr_spmv_inode_T<USECACHE,16,  32>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  64)   { csr_spmv_inode_T<USECACHE,16,  64>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  128)  { csr_spmv_inode_T<USECACHE,16, 128>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  256)  { csr_spmv_inode_T<USECACHE,16, 256>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  512)  { csr_spmv_inode_T<USECACHE,16, 512>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  1024) { csr_spmv_inode_T<USECACHE,16,1024>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { csr_spmv_inode_T<USECACHE,16,2048>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#endif
  }
  else {
    if (nnzPerRowMax <=  32)   { csr_spmv_inode_T<USECACHE,32,  32>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  64)   { csr_spmv_inode_T<USECACHE,32,  64>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  128)  { csr_spmv_inode_T<USECACHE,32, 128>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  256)  { csr_spmv_inode_T<USECACHE,32, 256>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  512)  { csr_spmv_inode_T<USECACHE,32, 512>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
    if (nnzPerRowMax <=  1024) { csr_spmv_inode_T<USECACHE,32,1024>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { csr_spmv_inode_T<USECACHE,32,2048>(*A, numNodes, nodeMax, inodes, x, y); return 0;}
#endif
  }
  
  //std::cout << "nnzPerRowMax exceeds limit in csr_spmv_inode" << std::endl;
  return 1;
}


#endif // CSR_SPMV_INODE_H
